// pi_user user space library for controlling PI stages via their C-843 PCI card
// under LGPL.
// You are free to copy and modify this code provided it remains under LGPL 
// and you do not charge for it.
// 
// You can use this library in your own code without conferring and right or
// obligation over your code.
// Matt Clark Feb 2003

/* (c) Matt Clark Feb 2003
 *
 * You are free to copy, change and distribute this software BUT
 * you cannot charge for it (even if you change it).
 *
 * The use of this library in your code conveys no rights over
 * your code.  So, if you wish to charge for a program that uses
 * this library that is fine.  You will have to distribute the
 * library free of charge.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Matt.Clark@Nottingham.ac.uk
 */

// Revision history
// Release 0.62 - Allow multiple calls to pi_openboard and pi_closeboard
// Release 0.61 - removal of pi_reset_board from pi_open_board
// Release 0.60 - new feature: pi_set_pos, 
// Release 0.54 - new feature: PI_[sg]et_vel_cps
//                new feature: PI_stop_dead	 August 2003
//
// Release 0.53 - new feature: PI_get_limit_status August 2003
// Release 0.52 - general beta release July 2003
#include "pi_user.h"
#include <math.h>

int pi_device;
// get version is sometimes realllly slow so
// cache it and the number of axis
int pi_n_axis=0; 
int pi_board_version=0;

BOOL pi_board_open=FALSE;

// global variables -
unsigned char	pi_digital_out=0,pi_motors_and_brakes=0;


#define CA(a,b)	(a|((unsigned short)b<<8))

// really just a place holder for housekeeping
// allow multiple calls
BOOL pi_closeboard(void){
	if(pi_board_open==TRUE){
		pi_board_open=FALSE;
		close(pi_device);
		}
	return TRUE;
	}

BOOL pi_openboard(void){
	return pi_openboard(PI_DEFAULT_DEVICE_NAME);
	}

// should implement some form of locking here
// although this will be done in the driver
// Allow this to be called many times if it is
// already open return TRUE as if it were succesfully
// opened.
BOOL pi_openboard(char * dev){
	if(pi_board_open==FALSE){
		pi_device=open(dev,O_RDWR);
		if(pi_device==-1){
			PDEBUG("pi_openboard: couldn't open the board\n\r");
			return FALSE;
			}
		if(pi_init_board()==TRUE){
			pi_board_open=TRUE;
			return TRUE;
			}
		return FALSE;
		}
	return TRUE;
	}
	

BOOL pi_select_board(BYTE board){
	PDEBUG("pi_select_board: this is not supported\n\r");
	PDEBUG("pi_select_board: the driver only supports one board\n\r");
	return FALSE;
	}


int pi_getnumberofinstalledboards(BYTE* pnumber){
	PDEBUG("pi_getnumberofinstalledboards: this is not supported\n\r");
	PDEBUG("pi_select_board: the driver only supports one board\n\r");
	return 1;
	}

int pi_get_version(void){
	if(pi_board_version==0){
		PI_DATA data[4];
		pi_io_crr(PI_GET_VERSION,data);
		pi_board_version=(data[1]<<16)+data[2];
		PDEBUG("pi_axis_installed: got %x and %x which makes %x \n\r",data[1],data[2],pi_board_version);
		}
	return pi_board_version;
	}

int pi_axis_installed(void){
int version;
	// read the number of axis on the board
	// request the version number then parse this 
	// for the number of installed axis
	if(pi_n_axis==0){
		version=pi_get_version();
		pi_n_axis=(int)((version>>16)&PI_AXIS_BITS)>>PI_AXIS_SHIFT_BITS;  // 16 because it has the high 2 bytes
		}
	return pi_n_axis;
	}

// following PI's example - may tidy later matt
BOOL pi_init_board(void){
int axis,i;
	// removed at version 0.61
	// removal allows position saving and recall
	// pi_reset_board();

	axis=pi_axis_installed();
	if((axis != 2) && (axis !=4)){
		PDEBUG("pi_openboard: illegal number of axis (%d)\n\r",axis);
		return FALSE;
		}
	for(i=0;i<axis;i++){
		PDEBUG("Initing axis %d\n\r",i);
		pi_io_cw(CA(PI_SET_MOTOR_MODE,i),1);
		pi_io_cw(CA(PI_SET_OUTPUT_MODE,i),0);
		pi_io_cw(CA(PI_SET_SAMPLE_TIME,i),410);
		pi_init_ls(i,PI_LIMIT_ON);
		}	

	pi_set_motor(0,PI_MOTORS_ON);
	pi_set_brake(0,PI_BRAKES_ON);
	pi_DO_set(0); // stop gap - should be in driver

	return TRUE;	
	}	

// supports modes PI_LIMIT_ON 1 and PI_LIMIT_OFF 0
void pi_init_ls(PI_AXIS axis,int mode){
	pi_io_cw(CA(PI_SET_LIMIT_SWITCH_MODE,axis),mode?1:0);
	}

	// sets the limit switch sense for axis 
int pi_auto_set_limit(PI_AXIS axis){
	PI_DATA data[4];
	// set the limit sens one way
	pi_io_cw(CA(PI_SET_SIGNAL_SENSE,axis),0);//usleep(1000000);// the windows driver sleeps 5 seconds - appears to make no difference - matt
	// read the limit status
	pi_io_cr(CA(PI_GET_SIGNAL_STATUS,axis),data);
	// these return data[0]=cms,data[1]=cmd
	// so we have three possibilities
	// all clear - means we have the correct limit switch sense
	// pos OR neg set - means we are on a limit switch and cannot decide
	// pos AND neg set - means we have go the sense wrong
	//
	// the limit switches from out point of view are active LOW hence
	// the screwed up logic here - matt Feb 2003
	if((data[1] & PI_SIGNAL_POS_LIMIT) && (data[1] & PI_SIGNAL_NEG_LIMIT)){
		PDEBUG("pi_auto_set_limit: tested axis %d and it is all clear\n\r",axis);
		PDEBUG("pi_auto_set_limit: limit sense is 0\n\r");
		pi_io_cw(CA(PI_SET_SIGNAL_SENSE,axis),0);
		return PI_LIMIT_ZERO;
		}
	if(!(data[1] & PI_SIGNAL_POS_LIMIT) && !(data[1] & PI_SIGNAL_NEG_LIMIT)){
		PDEBUG("pi_auto_set_limit: tested axis %d and both +ve and -ve are set\n\r",axis);
		PDEBUG("pi_auto_set_limit: limit sense is 1\n\r");
		pi_io_cw(CA(PI_SET_SIGNAL_SENSE,axis),PI_SIGNAL_POS_LIMIT|PI_SIGNAL_NEG_LIMIT|PI_ENCODER_INDEX|PI_ENCODER_HOME);
		//pi_io_cw(CA(PI_RESET_EVENT_STATUS,axis),~(PI_EVENT_POS_LIMIT | PI_EVENT_NEG_LIMIT));
		return PI_LIMIT_ONE;
		}
	PDEBUG("pi_auto_set_limit: one limit switch is activated for axis %d\n\r",axis);
	PDEBUG("pi_auto_set_limit: cannot autoset limit sense\n\r");
	return PI_LIMIT_FAIL;
	}	

void pi_set_limit(PI_AXIS axis,int sense){
	pi_io_cw(CA(PI_SET_SIGNAL_SENSE,axis),sense);
	}

int pi_get_limit_status(PI_AXIS axis){
int signal;
	signal=pi_get_signal_status(axis);
        if(!(signal & PI_SIGNAL_POS_LIMIT))
                return PI_ON_POSITIVE_LIMIT;
        if(!(signal & PI_SIGNAL_NEG_LIMIT))
                return PI_ON_NEGATIVE_LIMIT;
        return PI_LIMIT_OK;	
	}

int pi_get_limit_status(void){
int i;

	for(i=0;i<pi_n_axis;i++){
		if(pi_get_limit_status(i)!=PI_LIMIT_OK)
			return PI_LIMIT_BAD;
		}
	return PI_LIMIT_OK;
	}
	
// init each axis with default values following PI's example
BOOL pi_init_axis(PI_AXIS axis){
	pi_setQMC(PI_SET_AUTO_STOP_MODE,	axis,	1);
	pi_setQMC(PI_SET_POSITION_ERROR_LIMIT,	axis,	2000);
	pi_setQMC(PI_SET_KP,			axis,	120);
	pi_setQMC(PI_SET_KI,			axis,	100);
	pi_setQMC(PI_SET_KD,			axis,	100);
	pi_setQMC(PI_SET_VELOCITY,		axis,	50000);
	pi_setQMC(PI_SET_ACCELERATION,		axis,	120);
	pi_setQMC(PI_SET_DECELERATION,		axis,	120);
	pi_setQMC(PI_SET_KVFF,			axis,	2500);
	pi_setQMC(PI_SET_INTEGRATION_LIMIT,	axis,	32000);//32000 in the dll, 2000 recommended for the stages
	pi_setQMC(PI_RESET_EVENT_STATUS,	axis, 	0);
	pi_setQMC(PI_MULTI_UPDATE,		0,	1<<axis);
	pi_set_brake(axis,PI_BRAKES_ON);// enables all brakes	
	return TRUE;
	}

void pi_DO_up(unsigned char bit){
	pi_DO_set(pi_digital_out | bit);
	}

void pi_DO_down(unsigned char bit){
	pi_DO_set(pi_digital_out & ~bit);
	}

void pi_DO_set(unsigned char data){
	pi_digital_out=data;
	pi_io_cww(PI_WRITE_IO,0,pi_digital_out);
	}

unsigned char pi_n2b(int n){
	return 1 << n;
	}

int pi_get_DI(){
	return pi_getQMC(PI_READ_IO,0);
	}

void pi_set_brake(PI_AXIS axis, int state){
	switch(state){
		case PI_BRAKE_ON:
			pi_motors_and_brakes|=1<<(axis+4);
			break;
		case PI_BRAKE_OFF:
			pi_motors_and_brakes&=~(1<<(axis+4));
			break;
		case PI_BRAKES_ON:
			pi_motors_and_brakes|=0xf0;
			break;
		case PI_BRAKES_OFF:
			pi_motors_and_brakes&=0x0f;
			break;
			}
	pi_io_cww(PI_WRITE_IO,1,pi_motors_and_brakes);
	}


void pi_set_motor(PI_AXIS axis, int state){
	switch(state){
		case PI_MOTOR_ON:
			pi_motors_and_brakes|=1<<axis;
			PDEBUG("PI_MOTOR_ON\n\r");
			break;
		case PI_MOTOR_OFF:
			pi_motors_and_brakes&=~(1<<axis);
			PDEBUG("PI_MOTOR_OFF\n\r");
			break;
		case PI_MOTORS_ON:
			pi_motors_and_brakes|=0x0f;
			PDEBUG("PI_MOTORS_ON\n\r");
			break;
		case PI_MOTORS_OFF:
			pi_motors_and_brakes&=0xf0;
			PDEBUG("PI_MOTORS_OFF\n\r");
			break;
			}
	pi_io_cww(PI_WRITE_IO,1,pi_motors_and_brakes);
	}

int pi_motor_and_brake_status(void){
	//return pi_getQMC(PI_READ_IO,0,1);
	return pi_motors_and_brakes;
	}

int pi_get_motor(PI_AXIS axis){
	return (pi_motors_and_brakes&(1<<axis))?PI_MOTOR_ON:PI_MOTOR_OFF;
	}

int pi_get_brake(PI_AXIS axis){
	return (pi_motors_and_brakes&(1<<(axis+4)))?PI_BRAKE_ON:PI_BRAKE_OFF;
	}

// CAUTION CAUTION BYTE ORDERING NON X86
// int is a 32 bit number on all platforms
// driver expects little endian arguments (I know the driver should use __cpu_to_le16...)
void pi_set_pos(PI_AXIS axis, int position){
__u16 hi,low;
	hi = (position >> 16) & 0xffff;
	low= position & 0xffff;
	pi_io_cww(CA(PI_SET_ACTUAL_POSITION,axis),hi,low);
	}
	
void pi_wait_stop(PI_AXIS axis){
	// this <should> be done in the driver but we are waiting for the
	// stages to arrive before I can generate GPBINT interrupts
	// until then the cheese polling loop
	//
	// then I may put it in the driver and have an ioctl return
	// when done (if the interrupts can be configured to do-the-right-thing)
	do{ 
		usleep(50000);
		}
		while(pi_moving(axis)==TRUE);
	return;
	}

// returns TRUE if the axis is moving
BOOL pi_moving(PI_AXIS axis){
PI_DATA data[4];
	pi_io_cr(CA(PI_GET_ACTIVITY_STATUS,axis),data);
	return (data[1]&PI_IN_MOTION)? TRUE:FALSE;
	}

// returns the position 
int pi_get_pos(PI_AXIS axis){
PI_DATA data[4];
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis),data);
	return (data[1]<< 16) + data[2];
	}

// returns the position error 
int pi_get_pos_err(PI_AXIS axis){
PI_DATA data[4];
	pi_io_crr(CA(PI_GET_POSITION_ERROR,axis),data);
	return (data[1]<< 16) + data[2];
	}

// writes up to the number of axis supported into the array
// it isn't a very good function.
void pi_get_pos4(int *pos){
int axis,i;
	axis=pi_axis_installed();
	for(i=0;i<axis;i++){
		pos[i]=pi_get_pos(i);
		}
	}
	

// the 65536 scaling comes from the chipset
// the 2500 scaling comes (I think) from the sampling time
// and together they convert from velocity/interval to counts per second - matt
// following the poorly documented example in the pi dll
int pi_get_v(PI_AXIS axis){
PI_DATA data[4];
	pi_io_crr(CA(PI_GET_VELOCITY,axis),data);
	return (int)((float)((data[1]<< 16) + data[2])*2500.0/65536.0);
	}
// pi_get_vel_cps and pi_set_vel_cps should get and set the velocity in cps
int pi_get_vel_cps(PI_AXIS axis){
int     vel;
        vel=pi_getQMC(PI_GET_VELOCITY,axis);
        return (int)((float)vel*2500.0/65536.0);
        }

void pi_set_vel_cps(PI_AXIS axis, int vel){
        vel=(int)((float)vel*65536.0/2500.0);
        pi_setQMC(PI_SET_VELOCITY,axis,vel);
        }



int pi_get_activity_status(PI_AXIS axis){
PI_DATA data[4];
	pi_io_cr(CA(PI_GET_ACTIVITY_STATUS,axis),data);
	return data[1];
	}
	
int pi_get_event_status(PI_AXIS axis){
PI_DATA data[4];
	pi_io_cr(CA(PI_GET_EVENT_STATUS,axis),data);
	return data[1];
	}
	
	
int pi_get_signal_status(PI_AXIS axis){
PI_DATA data[4];
	pi_io_cr(CA(PI_GET_SIGNAL_STATUS,axis),data);
	return data[1];
	}

// I don't know what this function is for - matt
BOOL pi_get_refsignal(PI_AXIS axis){
PI_DATA data[4];
	pi_io_cr(CA(PI_GET_SIGNAL_STATUS,axis),data);
	return (data[1]&PI_ENCODER_INDEX)? TRUE:FALSE;
	}

// **********************************************************
// MOVE - movng functions

void pi_move_abs(PI_AXIS axis, int position){
	pi_io_cww(CA(PI_SET_POSITION,axis),(position>>16)&0xffff,(position)&0xffff);
	pi_io_cw(PI_MULTI_UPDATE,1<<axis);
	}


void pi_move_abs(PI_AXIS axis1, int position1, PI_AXIS axis2, int position2){
	pi_io_cww(CA(PI_SET_POSITION,axis1),(position1>>16)&0xffff,(position1)&0xffff);
	pi_io_cww(CA(PI_SET_POSITION,axis2),(position2>>16)&0xffff,(position2)&0xffff);
	pi_io_cw(PI_MULTI_UPDATE,(1<<axis1)|(1<<axis2));
	}

void pi_move_abs(PI_AXIS axis1, int position1, PI_AXIS axis2, int position2, PI_AXIS axis3, int position3){
	pi_io_cww(CA(PI_SET_POSITION,axis1),(position1>>16)&0xffff,(position1)&0xffff);
	pi_io_cww(CA(PI_SET_POSITION,axis2),(position2>>16)&0xffff,(position2)&0xffff);
	pi_io_cww(CA(PI_SET_POSITION,axis3),(position3>>16)&0xffff,(position3)&0xffff);
	pi_io_cw(PI_MULTI_UPDATE,(1<<axis1)|(1<<axis2)|(1<<axis3));
	}

void pi_move_abs(PI_AXIS axis1, int position1, PI_AXIS axis2, int position2, PI_AXIS axis3, int position3, PI_AXIS axis4, int position4){
	pi_io_cww(CA(PI_SET_POSITION,axis1),(position1>>16)&0xffff,(position1)&0xffff);
	pi_io_cww(CA(PI_SET_POSITION,axis2),(position2>>16)&0xffff,(position2)&0xffff);
	pi_io_cww(CA(PI_SET_POSITION,axis3),(position3>>16)&0xffff,(position3)&0xffff);
	pi_io_cww(CA(PI_SET_POSITION,axis4),(position4>>16)&0xffff,(position4)&0xffff);
	pi_io_cw(PI_MULTI_UPDATE,(1<<axis1)|(1<<axis2)|(1<<axis3)|(1<<axis4));
	}

void pi_move_rel(PI_AXIS axis, int position){
PI_DATA	data[4];
int	pos;
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis),data);
	pos=data[1]<<16+data[2]+position;
	pi_io_cww(CA(PI_SET_POSITION,axis),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_cw(PI_MULTI_UPDATE,1<<axis);
	}

void pi_move_rel(PI_AXIS axis1, int position1, PI_AXIS axis2, int position2){
PI_DATA	data[4];
int	pos;
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis1),data);
	pos=data[1]<<16+data[2]+position1;
	pi_io_cww(CA(PI_SET_POSITION,axis1),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis2),data);
	pos=data[1]<<16+data[2]+position2;
	pi_io_cww(CA(PI_SET_POSITION,axis2),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_cw(PI_MULTI_UPDATE,(1<<axis1)|(1<<axis2));
	}

void pi_move_rel(PI_AXIS axis1, int position1, PI_AXIS axis2, int position2, PI_AXIS axis3, int position3){
PI_DATA	data[4];
int	pos;
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis1),data);
	pos=data[1]<<16+data[2]+position1;
	pi_io_cww(CA(PI_SET_POSITION,axis1),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis2),data);
	pos=data[1]<<16+data[2]+position2;
	pi_io_cww(CA(PI_SET_POSITION,axis2),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis3),data);
	pos=data[1]<<16+data[2]+position3;
	pi_io_cww(CA(PI_SET_POSITION,axis3),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_cw(PI_MULTI_UPDATE,(1<<axis1)|(1<<axis2)|(1<<axis3));
	}

void pi_move_rel(PI_AXIS axis1, int position1, PI_AXIS axis2, int position2, PI_AXIS axis3, int position3, PI_AXIS axis4, int position4){
PI_DATA	data[4];
int	pos;
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis1),data);
	pos=data[1]<<16+data[2]+position1;
	pi_io_cww(CA(PI_SET_POSITION,axis1),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis2),data);
	pos=data[1]<<16+data[2]+position2;
	pi_io_cww(CA(PI_SET_POSITION,axis2),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis3),data);
	pos=data[1]<<16+data[2]+position3;
	pi_io_cww(CA(PI_SET_POSITION,axis3),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_crr(CA(PI_GET_ACTUAL_POSITION,axis4),data);
	pos=data[1]<<16+data[2]+position4;
	pi_io_cww(CA(PI_SET_POSITION,axis4),(pos>>16)&0xffff,(pos)&0xffff);
	pi_io_cw(PI_MULTI_UPDATE,(1<<axis1)|(1<<axis2)|(1<<axis3)|(1<<axis4));
	}


// I don't implement vector because of the unlinked acceleration profiles
// however (april 2003) PI have requested this so here goes (following their
// code)

void pi_vector_abs(PI_AXIS axis1, int p1, PI_AXIS axis2, int p2, int velocity){
int length;
int v1,v2;
int cp1,cp2;
int dp1,dp2;

	cp1=pi_getQMC(PI_GET_VELOCITY,axis1);
	cp2=pi_getQMC(PI_GET_VELOCITY,axis2);

	dp1=p1-cp1;
	dp2=p2-cp2;

	length=(int)sqrt((float)dp1*(float)dp1+(float)dp2*(float)dp2);
	v1=velocity*dp1/length;if(v1<0)v1=-v1;
	v2=velocity*dp2/length;if(v2<0)v2=-v2;
	// then get the current velocity and save it
	// then work out the shift and appropriate
	pi_setQMC(PI_SET_VELOCITY,axis1,v1);
	pi_setQMC(PI_SET_VELOCITY,axis2,v2);
	pi_move_abs(axis1, p1, axis2, p2);
	}

// also remember that this assumes the SAME stages.....
// we can't restore the velocity because we return immediately
void pi_vector_rel(PI_AXIS axis1, int p1, PI_AXIS axis2, int p2, int velocity){
int length;
int v1,v2;
	length=(int)sqrt((float)p1*(float)p1+(float)p2*(float)p2);
	v1=velocity*p1/length;if(v1<0)v1=-v1;
	v2=velocity*p2/length;if(v2<0)v2=-v2;
	// then get the current velocity and save it
	// then work out the shift and appropriate
	pi_setQMC(PI_SET_VELOCITY,axis1,v1);
	pi_setQMC(PI_SET_VELOCITY,axis2,v2);
	pi_move_rel(axis1, p1, axis2, p2);
	}



// END MOVE - end of the moving functions
// **********************************************************


int pi_find_home(PI_AXIS axis){
int dest,pos;
int sig_before,sig_after;
	// get our status assuming that the limit switches are OK
	sig_before=pi_get_signal_status(axis)&(PI_ENCODER_INDEX|PI_SIGNAL_POS_LIMIT|PI_SIGNAL_NEG_LIMIT);
	// direction needed to travel is -ve to find home
	if((sig_before&PI_ENCODER_INDEX)== 0)
		dest=PI_MOST_NEGATIVE;
	else
		dest=PI_MOST_POSITIVE;
	// now..... move and wait
	pi_move_abs(axis,dest);
	do{
		usleep(PI_MONITOR_WAIT);//10ms?	
		sig_after=pi_get_signal_status(axis)&(PI_ENCODER_INDEX|PI_SIGNAL_POS_LIMIT|PI_SIGNAL_NEG_LIMIT);
		}
		while(sig_after==sig_before);
	// now something must have changed....
	// 1st Stop!
	pos=pi_get_pos(axis);
	pi_move_abs(axis,pos);
	if((sig_after & PI_ENCODER_INDEX) == (sig_before & PI_ENCODER_INDEX)){
		// return an error
		return PI_LIMIT_BAD;
		}
	
	// now we always find the edge in the +ve direction so if we
	// the encoder is LOW (ie we are +ve) step backwards
	if((sig_after & PI_ENCODER_INDEX) == 0){
		pi_move_abs(axis,pos-PI_FIND_HOME_SHIFT);
		}
	pi_wait_stop(axis);	
	// now find it in the +ve direction
	// first move in +ve
	sig_before=pi_get_signal_status(axis)&(PI_ENCODER_INDEX|PI_SIGNAL_POS_LIMIT|PI_SIGNAL_NEG_LIMIT);
	pi_move_abs(axis,pos+PI_FIND_HOME_SHIFT);
	do{
		usleep(PI_MONITOR_WAIT/10);//1ms?
		sig_after=pi_get_signal_status(axis)&(PI_ENCODER_INDEX|PI_SIGNAL_POS_LIMIT|PI_SIGNAL_NEG_LIMIT);
		}
		while(sig_after==sig_before);
	// now stop!
	pos=pi_get_pos(axis);
	pi_move_abs(axis,pos);
	return PI_OK;
	}
	
void pi_stop_dead(PI_AXIS axis){
int pos;
	// now stop!
	pos=pi_get_pos(axis);
	pi_move_abs(axis,pos);
	}

void pi_stop_dead(void){
int i;
	for(i=0;i<pi_axis_installed();i++){
		pi_stop_dead(i);
		}
	}

// recover - tricky one here base on __move_off_limits
void pi_move_off_limit_switch(PI_AXIS axis){
PI_DATA data[4];
int pos,direction=0;
	// get the signal status 
	pi_io_cr(CA(PI_GET_SIGNAL_STATUS,axis),data);
	// check we were on a limit switch anyway!
	// remember ! active low
	if((data[1]&PI_SIGNAL_POS_LIMIT) && (data[1]&PI_SIGNAL_NEG_LIMIT))
		return;
		
/*	pre 0.51 version
 	if(!(data[1] & PI_SIGNAL_POS_LIMIT))
		direction=-PI_MOVE_OFF_LIMITS;
	
	if(!(data[1] & PI_SIGNAL_NEG_LIMIT))
		direction=PI_MOVE_OFF_LIMITS;
*/
// post 0.51 version
 	if(!(data[1] & PI_SIGNAL_POS_LIMIT))
		direction=PI_MOST_NEGATIVE;
	
	if(!(data[1] & PI_SIGNAL_NEG_LIMIT))
		direction=PI_MOST_POSITIVE;



	
	// disable the limit switches (in case we are on an edge an generate spurious events)
	pi_io_cw(CA(PI_SET_LIMIT_SWITCH_MODE,axis),0);

	// reset the limit switch events
	// CAUTION CHECK THE BIT FIELD / TYPE HERE FOR DODGINESS - matt feb 2003
	pi_io_cw(CA(PI_RESET_EVENT_STATUS,axis),~(PI_EVENT_POS_LIMIT | PI_EVENT_NEG_LIMIT));
	pi_setQMC(PI_SET_VELOCITY,axis,PI_DEF_VEL);
	// start to move away from the switch
/*	pre 0.51 version
 	pos=pi_get_pos(axis);
	pi_move_abs(axis,pos+direction);
*/
// post 0.51 version
	pi_move_abs(axis,direction);
// can't use the limit switches so wait until it has stopped	
	usleep(PI_MONITOR_WAIT);
/* pre 0.51 version
	do{
		usleep(PI_MONITOR_WAIT);
		}
		while(pi_moving(axis));
*/
// post 0.51 version
	do{
		usleep(PI_MONITOR_WAIT);
		pi_io_cr(CA(PI_GET_SIGNAL_STATUS,axis),data);
		}
		while(!(data[1]&PI_SIGNAL_POS_LIMIT) || !(data[1]&PI_SIGNAL_NEG_LIMIT));
	
	// we are now off the limit switch
	pos=pi_get_pos(axis);
	pi_move_abs(axis,pos);
	pi_io_cw(CA(PI_SET_STOP_MODE,axis),2);
	pi_io_c(CA(PI_UPDATE,axis));
	
	// enable the limit switch	
	pi_io_cw(CA(PI_SET_LIMIT_SWITCH_MODE,axis),1);
	}

int pi_getQMC(int cmd, PI_AXIS axis){
PI_DATA	data[4];

	switch(cmd){
		case	PI_GET_CHECKSUM:
		case	PI_GET_HOST_IO_ERROR: 
		case	PI_GET_INTERRUPT_AXIS:
		case	PI_GET_TRACE_STATUS:
		case	PI_GET_DIAGNOSTIC_PORT_MODE:
		case	PI_GET_SAMPLE_TIME:
		case	PI_GET_SERIAL_PORT_MODE:
		case	PI_GET_TRACE_MODE:
		case	PI_GET_TRACE_PERIOD:
		case	PI_GET_TRACE_START:
		case	PI_GET_TRACE_STOP:
                	pi_io_cr(cmd,data);
			return data[1];
        		break;

		case	PI_GET_ACTIVITY_STATUS:
		case	PI_GET_CURRENT_MOTOR_COMMAND:
		case	PI_GET_DERIVATIVE:
		case	PI_GET_EVENT_STATUS:
		case	PI_GET_SIGNAL_STATUS:
		case	PI_GET_AUTO_STOP_MODE:
		case	PI_GET_AXIS_MODE:
		case	PI_GET_AXIS_OUT_SOURCE:
		case	PI_GET_CAPTURE_SOURCE:
		case	PI_GET_DERIVATIVE_TIME:
		case	PI_GET_ENCODER_MODULUS:
		case	PI_GET_ENCODER_SOURCE:
		case	PI_GET_GEAR_MASTER:
		case	PI_GET_INTERRUPT_MASK:
		case	PI_GET_KAFF:
		case	PI_GET_KD:
		case	PI_GET_KI:
		case	PI_GET_KOUT:
		case	PI_GET_KP:
		case	PI_GET_KVFF:
		case	PI_GET_LIMIT_SWITCH_MODE:
		case	PI_GET_MOTION_COMPLETE_MODE:
		case	PI_GET_MOTOR_BIAS:
		case	PI_GET_MOTOR_COMMAND:
		case	PI_GET_MOTOR_LIMIT:
		case	PI_GET_MOTOR_MODE:
		case	PI_GET_OUTPUT_MODE:
		case	PI_GET_PROFILE_MODE:
		case	PI_GET_SETTLE_TIME:
		case	PI_GET_SETTLE_WINDOW:
		case	PI_GET_SIGNAL_SENSE:
		case	PI_GET_STOP_MODE:
		case	PI_GET_TRACKING_WINDOW:
		        pi_io_cr(CA(cmd,axis),data);
			return data[1];
			break;

		case	PI_GET_TIME:          
		case	PI_GET_TRACE_COUNT:
		case	PI_GET_VERSION:
	                pi_io_crr(cmd,data);
			return (data[1]<<16)+data[2];
        	        break;

		case	PI_GET_ACTUAL_VELOCITY: 
		case	PI_GET_CAPTURE_VALUE:
		case	PI_GET_COMMANDED_ACCELERATION:
		case	PI_GET_COMMANDED_POSITION:
		case	PI_GET_COMMANDED_VELOCITY:
		case	PI_GET_INTEGRAL:
		case	PI_GET_POSITION_ERROR:
		case	PI_GET_ACCELERATION:
		case	PI_GET_DECELERATION:
		case	PI_GET_ACTUAL_POSITION:
		case	PI_GET_GEAR_RATIO:
		case	PI_GET_INTEGRATION_LIMIT:
		case	PI_GET_JERK:
		case	PI_GET_POSITION:
		case	PI_GET_POSITION_ERROR_LIMIT:
		case	PI_GET_VELOCITY:
        		pi_io_crr(CA(cmd,axis),data);
			return (data[1]<<16)+data[2];
        		break;

        	default:
			PDEBUG("pi_user: error: unknown pi_getQMC command\n\r");
                	return 0;
		}
	}

int pi_getQMCA(int cmd, PI_AXIS axis, int arg){
	return pi_getQMC(cmd,axis,arg);
	}

int pi_getQMC(int cmd, PI_AXIS axis, int arg){
PI_DATA	data[4];

	switch(cmd){
		case	PI_READ_ANALOG:
		case	PI_READ_IO:
		case	PI_GET_TRACE_VARIABLE:  
			pi_io_cwr(cmd,arg,data);
			return data[2];
			break;
		
		case	PI_GET_BREAKPOINT:     
		case	PI_GET_BUFFER_FUNCTION:
			pi_io_cwr(CA(cmd,axis),arg,data);
			return data[2];
			break;
			
		case	PI_GET_BUFFER_LENGTH: 
		case	PI_GET_BUFFER_READ_INDEX:
		case	PI_GET_BUFFER_START:
		case	PI_READ_BUFFER:
		case	PI_GET_BUFFER_WRITE_INDEX:
			pi_io_cwrr(cmd,arg,data);
			return data[2]<<16+data[3];
			break;
			
		case	PI_GET_BREAKPOINT_VALUE:
			pi_io_cwrr(CA(cmd,axis),arg,data);
			return data[2]<<16+data[3];
			break;

		default:
			PDEBUG("pi_user: error: unknown pi_getQMC command\n\r");
                	return 0;
		}
	}

void pi_setQMC(int cmd){

	switch(cmd){
		case	PI_RESET:
			pi_reset_board();
			break;
	
		case	PI_CLEAR_INTERRUPT:
		case	PI_NO_OPERATION:
			pi_io_c(cmd);
			break;
		default:
			PDEBUG("pi_user: pi_setQMC: unknown command\n\r");
			}
		}
	
void pi_setQMC(int cmd, PI_AXIS axis){

	switch(cmd){
		case	PI_RESET:
			pi_reset_board();
			break;
	
		case	PI_CLEAR_INTERRUPT:
		case	PI_NO_OPERATION:
			pi_io_c(cmd);
			break;

		case	PI_CLEAR_POSITION_ERROR:
		case	PI_UPDATE:
			pi_io_c(CA(cmd,axis));
			break;
		default:
			PDEBUG("pi_user: pi_setQMC: unknown command\n\r");
			}
		}
	

void pi_setQMC(int cmd, PI_AXIS axis, int arg){

	switch(cmd){
		case	PI_RESET:
			pi_reset_board();
			break;
	
		case	PI_CLEAR_INTERRUPT:
		case	PI_NO_OPERATION:
			pi_io_c(cmd);
			break;

		case	PI_CLEAR_POSITION_ERROR:
		case	PI_UPDATE:
			pi_io_c(CA(cmd,axis));
			break;
	
		case	PI_MULTI_UPDATE: 
		case	PI_SET_DIAGNOSTIC_PORT_MODE:
		case	PI_SET_SAMPLE_TIME:
		case	PI_SET_SERIAL_PORT_MODE:
		case	PI_SET_TRACE_MODE:
		case	PI_SET_TRACE_PERIOD:
		case	PI_SET_TRACE_START:
		case	PI_SET_TRACE_STOP:
			pi_io_cw(cmd,(arg & 0x0000fff));
			break;
	
		case	PI_RESET_EVENT_STATUS: // 16 bit prameters with axis
		case	PI_SET_AUTO_STOP_MODE:
		case	PI_SET_AXIS_MODE:
		case	PI_SET_AXIS_OUT_SOURCE:
		case	PI_SET_CAPTURE_SOURCE:
		case	PI_SET_DERIVATIVE_TIME:
		case	PI_SET_ENCODER_MODULUS:
		case	PI_SET_ENCODER_SOURCE:
		case	PI_SET_GEAR_MASTER:
		case	PI_SET_INTERRUPT_MASK:
		case	PI_SET_KAFF:
		case	PI_SET_KD:
		case	PI_SET_KI:
		case	PI_SET_KOUT:
		case	PI_SET_KP:
		case	PI_SET_KVFF:
		case	PI_SET_LIMIT_SWITCH_MODE:
		case	PI_SET_MOTION_COMPLETE_MODE:
		case	PI_SET_MOTOR_BIAS:
		case	PI_SET_MOTOR_COMMAND:
		case	PI_SET_MOTOR_LIMIT:
		case	PI_SET_MOTOR_MODE:
		case	PI_SET_OUTPUT_MODE:
		case	PI_SET_PROFILE_MODE:
		case	PI_SET_SETTLE_TIME:
		case	PI_SET_SETTLE_WINDOW:
		case	PI_SET_SIGNAL_SENSE:
		case	PI_SET_STOP_MODE:
		case	PI_SET_TRACKING_WINDOW:
			pi_io_cw(CA(cmd,axis),(arg & 0xfff));
			break;
			
		case	PI_ADJUST_ACTUAL_POSITION:
		case	PI_SET_ACCELERATION:   // 32 bit prameters with axis
		case	PI_SET_ACTUAL_POSITION:
		case	PI_SET_DECELERATION:
		case	PI_SET_GEAR_RATIO:
		case	PI_SET_INTEGRATION_LIMIT:
		case	PI_SET_JERK:
		case	PI_SET_POSITION:
		case	PI_SET_POSITION_ERROR_LIMIT:
		case	PI_SET_VELOCITY:
			pi_io_cww(CA(cmd,axis),(arg & 0xffff0000)>>16,(arg & 0x0000ffff));
			break;
		
		default:
			PDEBUG("pi_user: pi_setQMC: unknown command\n\r");
			}
		}

void pi_setQMCA(int cmd, PI_AXIS axis, int arg1, int arg2){
	pi_setQMC(cmd,axis,arg1,arg2);
	}

void pi_setQMC(int cmd,PI_AXIS axis, int arg1, int arg2){

	switch(cmd){
		case	PI_SET_TRACE_VARIABLE:  // 2 args without axis
		case	PI_WRITE_IO:
			pi_io_cww(cmd,arg1,(arg2 & 0x0000ffff));
			break;
	
		case	PI_SET_BREAKPOINT:      // 2 args with axis
		case	PI_SET_BUFFER_FUNCTION:
			pi_io_cww(CA(cmd,axis),arg1,(arg2 & 0x0000ffff));
			break;
	
		case	PI_SET_BUFFER_LENGTH:   // 2 args without axis (2nd arg 32 bit)
		case	PI_SET_BUFFER_READ_INDEX:
		case	PI_SET_BUFFER_START:
		case	PI_SET_BUFFER_WRITE_INDEX:
		case	PI_WRITE_BUFFER:
			pi_io_cwww(cmd,arg1,(arg2 & 0xffff0000)>>16, (arg2 & 0x0000ffff));
			break;
	
			// below - caution!
			// some values of the breakpoint are two 16 bit words!
		case	PI_SET_BREAKPOINT_VALUE: // 2 args with axis, (2nd arg 32 bit)
			pi_io_cwww(CA(cmd,axis),arg1,(arg2 & 0xffff0000)>>16, (arg2 & 0x0000ffff));
		break;
	
		default:
			PDEBUG("pi_user: pi_setQMC: unknown command\n\r");
		}
	}


void pi_setQMC(int cmd,PI_AXIS axis, int arg1, int arg2, int arg3){

	switch(cmd){
		case	PI_SET_BREAKPOINT_VALUE: // 3 16 bit args with axis
			pi_io_cwww(CA(cmd,axis),arg1,arg2,arg3);
		break;
	
		default:
			PDEBUG("pi_user: pi_setQMC: unknown command\n\r");
		}
	}




// ok so it looks like long hand but the type checking 
// makes for elegant capture of compile time errors
inline int pi_io_c(PI_CMD_ONLY cmd){
	//printf("CMD:PI_CMD_ONLY %x\n\r",cmd);	
	return ioctl(pi_device, PI_SET_0, &cmd);
	}

inline int pi_io_cw(PI_CMD_WRITE1 cmd, PI_DATA data1){
PI_DATA data[4];
	data[0]=cmd;
	data[1]=data1;
	//printf("CMD:PI_CMD_WRITE1 %x\n\r",cmd);	
	return ioctl(pi_device, PI_SET_1, data);
	}

inline int pi_io_cww(PI_CMD_WRITE2 cmd, PI_DATA data1, PI_DATA data2){
PI_DATA data[4];
	data[0]=cmd;
	data[1]=data1;
	data[2]=data2;
	//printf("CMD:PI_CMD_WRITE2 %x\n\r",cmd);	
	return ioctl(pi_device, PI_SET_2, data);
	}

inline int pi_io_cwww(PI_CMD_WRITE3 cmd, PI_DATA data1, PI_DATA data2, PI_DATA data3){
PI_DATA data[4];
	data[0]=cmd;
	data[1]=data1;
	data[2]=data2;
	data[3]=data3;
	//printf("CMD:PI_CMD_WRITE3 %x\n\r",cmd);	
	return ioctl(pi_device, PI_SET_3, data);
	}

inline int pi_io_cr(PI_CMD_READ1 cmd, PI_DATA *data){
	data[0]=cmd;
	data[1]=0;
	//printf("CMD:PI_CMD_READ1 %x\n\r",cmd);	
	return ioctl(pi_device, PI_GET_1, data);
	}

inline int pi_io_crr(PI_CMD_READ2 cmd, PI_DATA *data){
	data[0]=cmd;
	data[1]=0;
	data[2]=0;
	//printf("CMD:PI_CMD_READ2 %x\n\r",cmd);	
	return ioctl(pi_device, PI_GET_2, data);
	}

inline int pi_io_cwr(PI_CMD_WRITE_READ1 cmd, PI_DATA arg, PI_DATA *data){
	data[0]=cmd;
	data[1]=arg;
	data[2]=0;
	data[3]=0;
	//printf("CMD:PI_CMD_WRITE_READ1 %x\n\r",cmd);	
	return ioctl(pi_device, PI_SET_GET_1, data);
	}	

inline int pi_io_cwrr(PI_CMD_WRITE_READ2 cmd, PI_DATA arg, PI_DATA *data){
	data[0]=cmd;
	data[1]=arg;
	data[2]=0;
	data[3]=0;
	//printf("CMD:PI_CMD_WRITE_READ2 %x\n\r",cmd);	
	return ioctl(pi_device, PI_SET_GET_2, data);
	}	
// board reset requires a special sequence of events and is handled
// in the driver itself 
void pi_reset_board(void){
	//printf("CMD:RESET_BOARD\n\r");	
	ioctl(pi_device, PI_RESET_BOARD, NULL);
	}

int pi_read_async_events(struct pi_stage_event_t *data,int n){
	return read(pi_device,data,n*sizeof(struct pi_stage_event_t))/sizeof(struct pi_stage_event_t);
	}

// this allows async events to be captured a user function
void pi_enable_async_capture(void func(int)){
struct sigaction action;
	// first we have to install the appropriate handler
	action.sa_handler=func;
	sigemptyset(&(action.sa_mask));
	action.sa_flags = 0;
	action.sa_restorer = NULL;
	sigaction(SIGIO,&action,NULL);
	// then we assign our PID to ownership of the file
	fcntl(pi_device,F_SETOWN,getpid());	
	// then we request asynchronous notification
	fcntl(pi_device,F_SETFL,O_ASYNC);
	}

